#pragma once
#include "Vert.h"
#define _USE_MATH_DEFINES
#include <cmath>


static float gaussian(float x, float std)
{
	return 1/(sqrt(2*M_PI)*std) * exp(-x*x/(2*std*std));
}

static void normalize(float v[3])
{
	float l = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
	l = 1 / (float)sqrt(l);

	v[0] *= l;
	v[1] *= l;
	v[2] *= l;
}

static void normalize(Vert v)
{
	float l = v.x*v.x + v.y*v.y + v.z*v.z;
	l = 1 / (float)sqrt(l);

	v.x *= l;
	v.y *= l;
	v.z *= l;
}

static float dot(float a[3], float b[3])
{
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}


static void crossproduct(float a[3], float b[3], float res[3])
{
	res[0] = (a[1] * b[2] - a[2] * b[1]);
	res[1] = (a[2] * b[0] - a[0] * b[2]);
	res[2] = (a[0] * b[1] - a[1] * b[0]);
}

static void crossproduct(const Vert &a, const Vert &b, float res[3])
{
	res[0] = (a.y * b.z - a.z * b.y);
	res[1] = (a.z * b.x - a.x * b.z);
	res[2] = (a.x * b.y - a.y * b.x);
}

static float length(float v[3])
{
	return (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

static float length(const Vert &v)
{
	return (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
}

static void multiply(const float m[4][4], const Vert &v, Vert &res)
{
		res.x = m[0][0]*v.x + m[1][0]*v.y + m[2][0]*v.z + m[3][0]*v.w;
		res.y = m[0][1]*v.x + m[1][1]*v.y + m[2][1]*v.z + m[3][1]*v.w;
		res.z = m[0][2]*v.x + m[1][2]*v.y + m[2][2]*v.z + m[3][2]*v.w;
		res.w = m[0][3]*v.x + m[1][3]*v.y + m[2][3]*v.z + m[3][3]*v.w;
}
static void multiply(const float a[4][4], const float b[16], float res[4][4])
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
		{
			res[j][i] = 0;
			for (int k = 0; k < 4; k++)
				res[j][i] += a[k][i]*b[k+ 4*j];
		}
}

static void multiply(const float a[16], const float b[4], float res[4])
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 1; j++)
		{
			res[i] = 0;
			for (int k = 0; k < 4; k++)
				res[i] += a[k*4 + i]*b[k+ 4*j];
		}
}

static void multiplyT(const float a[16], const float b[4], float res[4])
{

	for (int j = 0; j < 1; j++)
	  for (int i = 0; i < 4; i++)
		{
			res[i] = 0;
			for (int k = 0; k < 4; k++)
				res[i] += a[k*4 + i]*b[k+ 4*j];
		}
}
