#include "GL/glut.h"
#include "GL/glui.h"
#include "Cg/cg.h"
#include "Cg/cgGL.h"

#include <math.h>
#include "Hair.h"
#include "Model.h"
#include "Math.cpp"

int	main_window;

// the camera info
float eye[3];
float lookat[3];
float up[3];


GLUI *glui;
GLUI_Rollout		*object_rollout;
GLUI_Rotation		*object_rotation;
GLUI_Translation	*object_xz_trans;
GLUI_Translation	*object_y_trans;


class transformGroup {
public:
	transformGroup() {
		live_object_rotation[0] = 1;
		live_object_rotation[5] = 1;
		live_object_rotation[10] = 1;
		live_object_rotation[15] = 1;

		live_object_xz_trans[0] = 0;
		live_object_xz_trans[1] = 0;
		live_object_y_trans = 0;
	}
	float live_object_rotation[16];
	float live_object_xz_trans[2];
	float live_object_y_trans;
};

transformGroup *currentTransforms;
transformGroup lightTransform;
transformGroup hairVolumneTransform;
transformGroup modelTransform;

Model *model;

float shellHeight = .04;
float finHeight = .07;
float finScalingFactor = 10;
float tex_scale=4;
int blendFins = 1;
int numShells = 32;
float finDensity = .02;
int draw_shell_box=0;
int draw_shells=1;
int draw_fin_box=0;
int draw_fins=1;
float skin_r=1;
float skin_g=.8;
float skin_b=.5;
float minShadow = .1;
int shellList;

int outputWindowSize[] = {1024, 1024};

bool firstTime = true;

const float LIGHTCOLOR[] = {1, .8, .7, 1};
const float zero[] = {0, 0, 0, 1};
float Ka[] = {.1, .1, .1, 1};
float Kd[] = {1, 1, 1, 1};
float Ks[] = {1, 1, .8, 1};
float Pd = 3;
float Ps = 32;

static GLuint	illuminationMapTex = 0,
				*shellTexs = NULL,
				finTex = 0;

const int HAIR_VOLUME_SIZE = 2;
const int HAIR_VOLUME_HEIGHT = 1;
int numHairs = 3000;

Hair *hairs = NULL;

//Cg data
static CGcontext Context = NULL;
static CGprogram vertexProgram = NULL, fragmentProgram = NULL;
static CGparameter ModelViewMatrixParam = NULL;
static CGparameter ModelViewMatrixITParam = NULL;
static CGparameter ProjectionMatrixParam = NULL;
static CGparameter TextureMatrixParam = NULL;
static CGparameter LightPositionParam = NULL;
static CGparameter EyePositionParam = NULL;
static CGparameter PreModelMatrixParam = NULL;

static CGparameter IlluminationMapParam = NULL;
static CGparameter TextureParam = NULL;

static CGprofile vertexProfile, fragmentProfile;


void myGlutReshape(int	x, int y);
void myGlutDisplay(	void );

void myGlutIdle(void)
{
	// make sure the main window is active
	if (glutGetWindow() != main_window)
		glutSetWindow(main_window);

	// if you have moving objects, you can do that here

	// just keep redrawing the scene over and over
//	glutPostRedisplay();
}

// mouse handling functions for the main window
// left mouse translates or picks, middle zooms, right rotates
// keep track of which button is down and where the last position was
int cur_button = -1;
int last_x;
int last_y;

// catch mouse up/down events
void myGlutMouse(int button, int state, int x, int y)
{
	if (state == GLUT_DOWN) {
		cur_button = button;

		if (cur_button == GLUT_LEFT_BUTTON) {
		}
	}
	else
	{
		if (button == cur_button) {
			cur_button = -1;
		}
	}

	last_x = x;
	last_y = y;
}

// catch mouse move events
void myGlutMotion(int x, int y)
{
	// the change in mouse position
	int dx = x-last_x;
	int dy = y-last_y;

	float scale, len, theta;
	float neye[3], neye2[3];
	float f[3], r[3], u[3];

	switch(cur_button)
	{
	case GLUT_LEFT_BUTTON:
		// translate
		f[0] = lookat[0] - eye[0];
		f[1] = lookat[1] - eye[1];
		f[2] = lookat[2] - eye[2];
		u[0] = 0;
		u[1] = 1;
		u[2] = 0;

		// scale the change by how far away we are
		scale = sqrt(length(f)) * 0.007;

		crossproduct(f, u, r);
		crossproduct(r, f, u);
		normalize(r);
		normalize(u);

		eye[0] += -r[0]*dx*scale + u[0]*dy*scale;
		eye[1] += -r[1]*dx*scale + u[1]*dy*scale;
		eye[2] += -r[2]*dx*scale + u[2]*dy*scale;

		lookat[0] += -r[0]*dx*scale + u[0]*dy*scale;
		lookat[1] += -r[1]*dx*scale + u[1]*dy*scale;
		lookat[2] += -r[2]*dx*scale + u[2]*dy*scale;

		break;

	case GLUT_MIDDLE_BUTTON:
		// zoom
		f[0] = lookat[0] - eye[0];
		f[1] = lookat[1] - eye[1];
		f[2] = lookat[2] - eye[2];

		len = length(f);
		normalize(f);

		// scale the change by how far away we are
		len -= sqrt(len)*-dy*0.01;

		eye[0] = lookat[0] - len*f[0];
		eye[1] = lookat[1] - len*f[1];
		eye[2] = lookat[2] - len*f[2];

		// make sure the eye and lookat points are sufficiently far away
		// push the lookat point forward if it is too close
		if (len < 1)
		{
			//printf("lookat move: %f\n", len);
			lookat[0] = eye[0] + f[0];
			lookat[1] = eye[1] + f[1];
			lookat[2] = eye[2] + f[2];
		}

		break;

	case GLUT_RIGHT_BUTTON:
		// rotate

		neye[0] = eye[0] - lookat[0];
		neye[1] = eye[1] - lookat[1];
		neye[2] = eye[2] - lookat[2];

		// first rotate in the x/z plane
		theta = -dx * 0.005;
		neye2[0] = (float)cos(theta)*neye[0] + (float)sin(theta)*neye[2];
		neye2[1] = neye[1];
		neye2[2] =-(float)sin(theta)*neye[0] + (float)cos(theta)*neye[2];


		// now rotate vertically
		theta = -dy * 0.005;

		f[0] = -neye2[0];
		f[1] = -neye2[1];
		f[2] = -neye2[2];
		u[0] = 0;
		u[1] = 1;
		u[2] = 0;
		crossproduct(f, u, r);
		crossproduct(r, f, u);
		len = length(f);
		normalize(f);
		normalize(u);

		neye[0] = len * ((float)cos(theta)*f[0] + (float)sin(theta)*u[0]);
		neye[1] = len * ((float)cos(theta)*f[1] + (float)sin(theta)*u[1]);
		neye[2] = len * ((float)cos(theta)*f[2] + (float)sin(theta)*u[2]);

		eye[0] = lookat[0] - neye[0];
		eye[1] = lookat[1] - neye[1];
		eye[2] = lookat[2] - neye[2];

		break;
	}


	last_x = x;
	last_y = y;

	glutPostRedisplay();
}

// you can put keyboard shortcuts in here
void myGlutKeyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
	// quit
	case 27: 
	case 'q':
	case 'Q':
		exit(0);
		break;
	}

	glutPostRedisplay();
}


// the window has changed shapes, fix ourselves up
// Note that some trickery goes on in here where x and y
// correspond to the picking point, not window size.

void myGlutReshape(int	x, int y)
{
	int viewport[4];
	GLUI_Master.get_viewport_area(&viewport[0], &viewport[1], 
								  &viewport[2], &viewport[3]);
	glViewport(viewport[0], viewport[1], 
			   viewport[2], viewport[3]);

	// projection transform
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	const float fieldOfView = 50;
	const float nearPlane = .01;
	const float farPlane = 1000;
	gluPerspective(fieldOfView, (double)viewport[2] / viewport[3], nearPlane, farPlane);
	glutPostRedisplay();
}

static void CheckCgError(CGprogram Program)
{
    CGerror err = cgGetError();

    if (err != CG_NO_ERROR)
    {
        printf("CG error: %s\n", cgGetErrorString(err));
		CGcontext context = cgGetProgramContext(Program);

		printf("Error listing follows:\n %s\n", cgGetLastListing(context));
        exit(1);
    }
}

void initializeCg() {

    if (cgGLIsProfileSupported(CG_PROFILE_VP20))
        vertexProfile = CG_PROFILE_VP20;
    else if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1))
        vertexProfile = CG_PROFILE_ARBVP1;
    else
    {
        fprintf(stderr, "Video card does not support vertex programs, exiting...\n");
        exit(-1);
    }

    /* Test cgContext creation */
    Context = cgCreateContext();
	CheckCgError(vertexProgram);

    /* Test adding source text to context */
    vertexProgram = cgCreateProgramFromFile(Context,
        CG_SOURCE, "" "lighting_vertex.cg",
        vertexProfile,
        NULL, NULL);
    CheckCgError(vertexProgram);

    fprintf(stderr, "LAST LISTING----%s----\n", cgGetLastListing(Context));

    fprintf(stderr, "---- PROGRAM BEGIN ----\n%s---- PROGRAM END ----\n",
        cgGetProgramString(vertexProgram, CG_COMPILED_PROGRAM));

    if(vertexProgram != NULL)
    {
        cgGLLoadProgram(vertexProgram);
        CheckCgError(vertexProgram);

        ModelViewMatrixParam = cgGetNamedParameter(vertexProgram, "ModelViewMatrix");
        CheckCgError(vertexProgram);
		
		ProjectionMatrixParam = cgGetNamedParameter(vertexProgram, "ProjectionMatrix");
        CheckCgError(vertexProgram); 

		ModelViewMatrixITParam = cgGetNamedParameter(vertexProgram, "ModelViewMatrixIT");
        CheckCgError(vertexProgram); 
		
		TextureMatrixParam = cgGetNamedParameter(vertexProgram, "TextureMatrix");
        CheckCgError(vertexProgram);
		
		LightPositionParam = cgGetNamedParameter(vertexProgram, "LightPosition");
		CheckCgError(vertexProgram);

		EyePositionParam = cgGetNamedParameter(vertexProgram, "EyePosition");
		CheckCgError(vertexProgram);

		PreModelMatrixParam = cgGetNamedParameter(vertexProgram, "PreModelMatrix");
		CheckCgError(vertexProgram);

    }

    if (cgGLIsProfileSupported(CG_PROFILE_FP20))
		fragmentProfile = CG_PROFILE_FP20;
    else if (cgGLIsProfileSupported(CG_PROFILE_ARBFP1))
        fragmentProfile = CG_PROFILE_ARBFP1;
    else
    {
        fprintf(stderr, "Video card does not support fragment programs, exiting...\n");
        exit(-1);
    }

    /* Test adding source text to context */
	fragmentProgram = cgCreateProgramFromFile(Context,
        CG_SOURCE, "" "lighting_fragment.cg",
        fragmentProfile,
        NULL, NULL);
    CheckCgError(fragmentProgram);

    fprintf(stderr, "LAST LISTING----%s----\n", cgGetLastListing(Context));

    fprintf(stderr, "---- PROGRAM BEGIN ----\n%s---- PROGRAM END ----\n",
        cgGetProgramString(fragmentProgram, CG_COMPILED_PROGRAM));

    if(fragmentProgram != NULL)
    {
        cgGLLoadProgram(fragmentProgram);
        CheckCgError(fragmentProgram);

		IlluminationMapParam = cgGetNamedParameter(fragmentProgram, "IlluminationMap");
		CheckCgError(fragmentProgram);

		TextureParam = cgGetNamedParameter(fragmentProgram, "hairTex");
		CheckCgError(fragmentProgram);

		cgGLSetTextureParameter(IlluminationMapParam, illuminationMapTex);
    }
}

const int TEX_SIZE = 128;
float illuminationData[TEX_SIZE+2][TEX_SIZE+2][4];
void makeIlluminationTex(int foo)
{
	for (int i = 0; i < TEX_SIZE+2; i++)
		for (int j = 0; j < TEX_SIZE+2; j++)
		{
			for (int k = 0; k  < 3; k++) { //ignore alpha here
				float u = (float)i/(TEX_SIZE+2);
				float v = (float)j/(TEX_SIZE+2);
				illuminationData[i][j][k] =  Ka[k] + Kd[k]* pow((1-u*u), Pd/2) + Ks[k]*pow((1-v*v), Ps/2);
			}
		}

	if(!illuminationMapTex)
		glGenTextures(1, &illuminationMapTex);
	glBindTexture(GL_TEXTURE_2D, illuminationMapTex);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEX_SIZE+2, TEX_SIZE+2, 1, GL_RGBA,
				 GL_FLOAT, illuminationData);
}

void makeHairGeometryVolume(int ignoreme)
{
	firstTime = true;
	if (hairs) 
		delete[] hairs;

	hairs = new Hair[numHairs];

	for (int i = 0; i < numHairs; i++) {
		hairs[i].GenerateHair(HAIR_VOLUME_HEIGHT, 
					Vert( ((float)rand())/RAND_MAX * HAIR_VOLUME_SIZE,
						   0,
						  ((float)rand())/RAND_MAX * HAIR_VOLUME_SIZE));
	}

	//make it torodial
	for (int i = 0; i < numHairs; i++) {
		bool invXPos = false, invXNeg = false,
			 invZPos = false, invZNeg = false;
		for (int k = 0; k < HAIR_RESOLUTION; k++) {
			hairs[i].flip[k] = false;

			if (hairs[i].vertices[k].x < 0) {
				hairs[i].vertices[k].x += HAIR_VOLUME_SIZE;
				if (!invXNeg)
					hairs[i].flip[k] = true;
				invXNeg = true;
			}
			else if (invXNeg) {
				invXNeg = false;
				hairs[i].flip[k] = true;
			}
			if (hairs[i].vertices[k].x > HAIR_VOLUME_SIZE) {
				hairs[i].vertices[k].x -= HAIR_VOLUME_SIZE;
				if (!invXPos)
					hairs[i].flip[k] = true;
				invXPos = true;
			}
			else if (invXPos) {
				invXPos = false;
				hairs[i].flip[k] = true;
			}
			if (hairs[i].vertices[k].z < 0) {
				hairs[i].vertices[k].z += HAIR_VOLUME_SIZE;
				if (!invZNeg)
					hairs[i].flip[k] = true;
				invZNeg = true;
			}
			else if (invZNeg) {
				invZNeg = false;
				hairs[i].flip[k] = true;
			}
			if (hairs[i].vertices[k].z > HAIR_VOLUME_SIZE) {
				hairs[i].vertices[k].z -= HAIR_VOLUME_SIZE;
				if (!invZPos)
					hairs[i].flip[k] = true;
				invZPos = true;
			}
			else if (invZPos) {
				invZPos = false;
				hairs[i].flip[k] = true;
			}
		}
	}
}

void drawHairGeometryVolume()
{
	glPushMatrix();
	glTranslatef(hairVolumneTransform.live_object_xz_trans[0], 
				 hairVolumneTransform.live_object_y_trans,
				 hairVolumneTransform.live_object_xz_trans[1]);
	glMultMatrixf((float*)hairVolumneTransform.live_object_rotation);

	for (int i = 0; i < numHairs; i++) {
		hairs[i].draw();
	}
	glPopMatrix();
}

void makeShell(int shellNum, int totalShells)
{
	//TODO: should probably use p-buffer so any size works.

	const int TEX_SIZE = 128;
	
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glViewport(0, 0, TEX_SIZE*HAIR_VOLUME_SIZE, TEX_SIZE*HAIR_VOLUME_SIZE);
	glOrtho(0, HAIR_VOLUME_SIZE, 
			0, HAIR_VOLUME_SIZE,
			-HAIR_VOLUME_HEIGHT*5, HAIR_VOLUME_HEIGHT*5);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glTranslatef(0, 0, -HAIR_VOLUME_HEIGHT*2);
	glRotatef(-90, 1, 0, 0);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask(0);

	if(shellNum==0)	//Draw an opaque rectangle over the viewport with skin color
	{
		glColor4f(skin_r,skin_g,skin_b,1);
		glBegin(GL_QUADS);
			glVertex3f(-HAIR_VOLUME_SIZE*2,-1,-HAIR_VOLUME_SIZE*2);
			glVertex3f(-HAIR_VOLUME_SIZE*2,-1,HAIR_VOLUME_SIZE*2);
			glVertex3f(HAIR_VOLUME_SIZE*2,-1,HAIR_VOLUME_SIZE*2);
			glVertex3f(HAIR_VOLUME_SIZE*2,-1,-HAIR_VOLUME_SIZE*2);
		glEnd();
	}

	for (int i = 0; i < numHairs; i++) {
		float alphas[HAIR_RESOLUTION];
		for (int j = 0; j < HAIR_RESOLUTION; j++) {
			float dist = hairs[i].vertices[j].y - (float)shellNum/numShells;
			float alpha_g = gaussian(dist*8, .4);
			alphas[j] = alpha_g;
		}
		hairs[i].draw(alphas);
	}

	glDepthMask(1);

	glEnable(GL_TEXTURE_2D);
	glGenTextures(1, &(shellTexs[shellNum]));
	glBindTexture(GL_TEXTURE_2D, (shellTexs[shellNum]));
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, TEX_SIZE*HAIR_VOLUME_SIZE, TEX_SIZE*HAIR_VOLUME_SIZE, 0);
	glDisable(GL_TEXTURE_2D);

	glPopMatrix();
	myGlutReshape(0,0);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void makeShells(int ignoreme)
{
	if (shellTexs) {
		glDeleteTextures(numShells, shellTexs);
		delete[] shellTexs;
	}

	shellTexs = new GLuint[numShells];
	glDisable(GL_LIGHTING);
	for (int i = 0; i < numShells; i++) {
		makeShell(i, numShells);
	}
	glEnable(GL_LIGHTING);
}

//density goes from [0,1]
void makeFinTex(float density)
{
	//TODO: should probably use p-buffer so any size works.

	float finTexSize = 256;

	float width = density * HAIR_VOLUME_SIZE;
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glViewport(0, 0, finTexSize*HAIR_VOLUME_SIZE, finTexSize*HAIR_VOLUME_HEIGHT);
	glOrtho(0, HAIR_VOLUME_SIZE, 
			0, HAIR_VOLUME_HEIGHT,
			-HAIR_VOLUME_SIZE*2, HAIR_VOLUME_SIZE*2);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_LIGHTING);
	for (int i = 0; i < numHairs; i++) {
		if (hairs[i].vertices[0].z < width)
			hairs[i].draw();
	}	
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	if (finTex) 
		glDeleteTextures(1, &finTex);
	glGenTextures(1, &finTex);
	glBindTexture(GL_TEXTURE_2D, finTex);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, finTexSize*HAIR_VOLUME_SIZE, finTexSize*HAIR_VOLUME_HEIGHT, 0);
	glDisable(GL_TEXTURE_2D);

	glPopMatrix();
	myGlutReshape(0,0);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)>(b))?(b):(a))
void drawFin(const Vert &v1, const Vert &v2, const Vert &N1, const Vert &N2, float eyeDir[3])
{
	float n1[3] = {N1.x*finHeight,
				   N1.y*finHeight,
				   N1.z*finHeight};
	float n2[3] = {N2.x*finHeight,
				   N2.y*finHeight,
				   N2.z*finHeight};

	float minShadow = 1;

	float finN[3];
	Vert n = v1 - v2;
	normalize(n);
	crossproduct(n, N1, finN);
	normalize(finN);
	float alpha = blendFins? max(0, 2*fabs(dot(finN, eyeDir)) - 1): 1;

	if (alpha == 0) return;

	float texStart = (float)rand()/RAND_MAX;

	float l = length(v1 - v2);

	glBegin(GL_QUADS);
	
	glColor4f(minShadow, 0, 0, alpha);
	glNormal3f(N1.x, N1.y, N1.z); 
	glTexCoord2f(texStart, 0); 
	glVertex3f(v1.x, v1.y, v1.z); 
	
	glColor4f(1, 0, 0, alpha);
	glTexCoord2f(texStart, 1); 
	glVertex3f(v1.x + n1[0], v1.y + n1[1], v1.z + n1[2]);

	
	glNormal3f(N2.x, N2.y, N2.z); 
	glTexCoord2f(texStart + finScalingFactor*l, 1); 
	glVertex3f(v2.x + n2[0], v2.y + n2[1], v2.z + n2[2]);
    
	glColor4f(minShadow, 0, 0, alpha); 
	glTexCoord2f(texStart + finScalingFactor*l, 0); 
	glVertex3f(v2.x, v2.y, v2.z);
	
	glEnd();
}

void makeModel()
{
	model = new Model();
	model->LoadModel("OBJ/bunny-1500-cleaned.obj");
}

void drawModel()
{
	cgGLSetStateMatrixParameter(PreModelMatrixParam,
		CG_GL_MODELVIEW_MATRIX,
		CG_GL_MATRIX_IDENTITY);

	glPushMatrix();
	glTranslatef(modelTransform.live_object_xz_trans[0], 
				 modelTransform.live_object_y_trans,
				 modelTransform.live_object_xz_trans[1]);
	glMultMatrixf((float*)modelTransform.live_object_rotation);

	cgGLBindProgram(vertexProgram);
	CheckCgError(vertexProgram);
	cgGLBindProgram(fragmentProgram);
	CheckCgError(fragmentProgram);

	cgGLSetStateMatrixParameter(ModelViewMatrixParam,
		CG_GL_MODELVIEW_MATRIX,
		CG_GL_MATRIX_IDENTITY);

	cgGLSetStateMatrixParameter(ProjectionMatrixParam,
		CG_GL_PROJECTION_MATRIX,
		CG_GL_MATRIX_IDENTITY);

	cgGLSetStateMatrixParameter(ModelViewMatrixITParam,
		CG_GL_MODELVIEW_MATRIX,
		CG_GL_MATRIX_INVERSE_TRANSPOSE);
	

	float lightPosition[] = {lightTransform.live_object_xz_trans[0],
						lightTransform.live_object_y_trans,
						lightTransform.live_object_xz_trans[1], 1};
	cgGLSetParameter4fv(LightPositionParam, lightPosition);

	cgGLSetParameter3fv(EyePositionParam, eye);

	cgGLEnableProfile(vertexProfile);
	cgGLEnableProfile(fragmentProfile);

	//cgGLEnableTextureParameter(TextureParam);
	cgGLEnableTextureParameter(IlluminationMapParam);


	if (true)//draw surface
	{
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glScalef(tex_scale,tex_scale,1);
		glMatrixMode(GL_MODELVIEW);

		cgGLSetStateMatrixParameter(TextureMatrixParam,
		CG_GL_TEXTURE_MATRIX,
		CG_GL_MATRIX_IDENTITY);

		glEnable(GL_TEXTURE_2D);
		
		float shadowPercent = minShadow;

		glColor4f(1, 0, 0, 1);
		glBindTexture(GL_TEXTURE_2D, shellTexs[0]);
		
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glDisable(GL_BLEND);
		glCallList(shellList);
		glEnable(GL_BLEND);	
		glDisable(GL_TEXTURE_2D);
		
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	}

	float eyeDirTemp[4] = {	eye[0] - modelTransform.live_object_xz_trans[0],
						eye[1] - modelTransform.live_object_y_trans,
						eye[2] - modelTransform.live_object_xz_trans[1]};
	float eyeDir[4];
	float transpose[16];
	//quick and ugly hack because we're lazy...
	transpose[0]=modelTransform.live_object_rotation[0];
	transpose[1]=modelTransform.live_object_rotation[4];
	transpose[2]=modelTransform.live_object_rotation[8];
	transpose[3]=modelTransform.live_object_rotation[12];
	transpose[4]=modelTransform.live_object_rotation[1];
	transpose[5]=modelTransform.live_object_rotation[5];
	transpose[6]=modelTransform.live_object_rotation[9];
	transpose[7]=modelTransform.live_object_rotation[13];
	transpose[8]=modelTransform.live_object_rotation[2];
	transpose[9]=modelTransform.live_object_rotation[6];
	transpose[10]=modelTransform.live_object_rotation[10];	
	transpose[11]=modelTransform.live_object_rotation[14];
	transpose[12]=modelTransform.live_object_rotation[3];
	transpose[13]=modelTransform.live_object_rotation[7];
	transpose[14]=modelTransform.live_object_rotation[11];
	transpose[15]=modelTransform.live_object_rotation[15];
	
	multiply(transpose, eyeDirTemp, eyeDir);
	normalize(eyeDir);

	if(draw_fins)
	{
		cgGLSetStateMatrixParameter(TextureMatrixParam,
		CG_GL_TEXTURE_MATRIX,
		CG_GL_MATRIX_IDENTITY);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, finTex);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glDepthMask(0);
		for (int i = 0; i < model->numtriangles; i++) {
			Vert normal[3] = {	model->triangles[i].normals[0],
								model->triangles[i].normals[1],
								model->triangles[i].normals[2]};
			
			srand(i); //needed for getting random fin tex in drawFin

			drawFin(model->triangles[i].v[0], 
					model->triangles[i].v[1],
					normal[0], normal[1], eyeDir);
			drawFin(model->triangles[i].v[1], 
					model->triangles[i].v[2],
					normal[1], normal[2], eyeDir);
			drawFin(model->triangles[i].v[2], 
					model->triangles[i].v[0],
					normal[2], normal[0],eyeDir);
		}
		//glDisable(GL_CULL_FACE);
		glDepthMask(1);

		glDisable(GL_TEXTURE_2D);
	}

	if(draw_shells)
	{
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glScalef(tex_scale,tex_scale,1);
		glMatrixMode(GL_MODELVIEW);

		cgGLSetStateMatrixParameter(TextureMatrixParam,
		CG_GL_TEXTURE_MATRIX,
		CG_GL_MATRIX_IDENTITY);

		glEnable(GL_TEXTURE_2D);
		glAlphaFunc(GL_GREATER,.1);
		glEnable(GL_ALPHA_TEST);
		for (int k = 1; k < numShells; k++) {

			double d=shellHeight/(numShells-1)*(k);	//Distance to translate triangle along normal
			
			float shadowPercent = min(1, 2*d/shellHeight*(1-minShadow) + minShadow);

			glColor4f(shadowPercent, d, 0, 1);
			glBindTexture(GL_TEXTURE_2D, shellTexs[k]);
			
			//cgGLSetTextureParameter(TextureParam, shellTexs[k]);
			
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

			glCallList(shellList);
		
		}
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_ALPHA_TEST);
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	}



	cgGLDisableTextureParameter(IlluminationMapParam);
	cgGLDisableProfile(vertexProfile);
	cgGLDisableProfile(fragmentProfile);

	glPopMatrix();
}


// draw the scene
void myGlutDisplay( void )
{
	glui->sync_live();
	glClearColor(1, 1, 1, 0);

	if (firstTime) {
		makeFinTex(finDensity);
		makeShells(numShells);
		firstTime = false;
	}

	
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	//glEnable(GL_NORMALIZE); //only needed if we are scaling...

	// camera transform
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], up[0], up[1], up[2]);
	
	float lightPosition[] = {lightTransform.live_object_xz_trans[0],
							lightTransform.live_object_y_trans,
							lightTransform.live_object_xz_trans[1], 1};
		
	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
	
	if(draw_fin_box)
	{
		drawHairGeometryVolume();
	}

	//glEnable(GL_TEXTURE_2D);
	////glBindTexture(GL_TEXTURE_2D, finTex);
	//glBindTexture(GL_TEXTURE_2D, finTex);
	//glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	//glBegin(GL_QUADS);
	//glTexCoord2f(0, 0); glVertex3f(0, 0, 0);
	//glTexCoord2f(0, 1); glVertex3f(0, 1, 0);
	//glTexCoord2f(1, 1); glVertex3f(2, 1, 0);
	//glTexCoord2f(1, 0); glVertex3f(2, 0, 0);
	//glEnd();
	//glDisable(GL_TEXTURE_2D);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if(draw_shell_box)
	{
		//draw example fin or shell
		glEnable(GL_TEXTURE_2D);
		//glBindTexture(GL_TEXTURE_2D, finTex);
		for (int i = 0; i < numShells; i++)
		{
			glBindTexture(GL_TEXTURE_2D, shellTexs[i]);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			glBegin(GL_QUADS);
			glTexCoord2f(0, 0); glVertex3f(0, 0, (float)i/numShells);
			glTexCoord2f(0, 1); glVertex3f(0, 1, (float)i/numShells);
			glTexCoord2f(1, 1); glVertex3f(1, 1, (float)i/numShells);
			glTexCoord2f(1, 0); glVertex3f(1, 0, (float)i/numShells);
			glEnd();
		}
		glDisable(GL_TEXTURE_2D);
	}

	drawModel();
	
	glDisable(GL_BLEND);
	//draw Light
	glPushAttrib(GL_LIGHTING_BIT);
	glPushMatrix();
	  glTranslatef(lightPosition[0], lightPosition[1], lightPosition[2]);
	  glMaterialfv(GL_FRONT, GL_EMISSION, LIGHTCOLOR);
	  glutSolidSphere(1, 5, 5);
	glPopMatrix();
	glPopAttrib();

	glutSwapBuffers(); 
}

transformGroup temp;
void switchTransform(int foo) 
{
	object_rotation->set_ptr_val(temp.live_object_rotation);
	object_rotation->set_float_array_val(currentTransforms->live_object_rotation);
	object_rotation->set_ptr_val(currentTransforms->live_object_rotation);
	
	object_xz_trans->set_ptr_val(temp.live_object_xz_trans);
	object_xz_trans->set_float_array_val(currentTransforms->live_object_xz_trans);
	object_xz_trans->set_ptr_val(currentTransforms->live_object_xz_trans);
	
	
	object_y_trans->set_ptr_val(&temp.live_object_y_trans);
	temp.live_object_y_trans = currentTransforms->live_object_y_trans;
	glui->sync_live();
	object_y_trans->set_float_val(currentTransforms->live_object_y_trans);
	
	object_y_trans->set_ptr_val(&currentTransforms->live_object_y_trans);
}

void makeHairUpdate(int type)
{
	makeFinTex(finDensity);
	if ((int) &finDensity != type)
		makeShells(0);
}

void makeDisplayList()
{
	shellList=glGenLists(1);
	glNewList(shellList,GL_COMPILE);

	glBegin(GL_TRIANGLES);
	
	for (int i = 0; i < model->numtriangles; i++) {
	     for (int j = 0; j < 3; j++)
	     {
			glTexCoord2f(T.x, T.y);
			glNormal3f(N.x, N.y, N.z);

			//translate along normal of triangle
	        glVertex3f(V.x, V.y, V.z);
             }
	 }
	 glEnd();

	 glEndList();
}
// entry point
int main(int argc, char* argv[])
{
	// initialize the camera
	eye[0] = 0;
	eye[1] = 0;
	eye[2] = 3;
	lookat[0] = 0;
	lookat[1] = 0;
	lookat[2] = 0;
	up[0] = 0;
	up[1] = 1;
	up[2] = 0;

	for (int i = 1; i < argc; i++) {
		printf("%s\n", argv[i]);
		if ( strcmp(argv[i], "-e") == 0) {
			eye[0] = atof(argv[++i]);
			eye[1] = atof(argv[++i]);
			eye[2] = atof(argv[++i]);
		} else if (strcmp(argv[i], "-a") == 0) {
			lookat[0] = atof(argv[++i]);
			lookat[1] = atof(argv[++i]);
			lookat[2] = atof(argv[++i]);
		} else if (strcmp(argv[i], "-u") == 0){
			up[0] = atof(argv[++i]);
			up[1] = atof(argv[++i]);
			up[2] = atof(argv[++i]);
		} else if (strcmp(argv[i], "-sx") == 0) {
			outputWindowSize[0] = atof(argv[++i]);
		} else if (strcmp(argv[i], "-sy") == 0) {
			outputWindowSize[1] = atof(argv[++i]);
		}
	}
	
	//
	// create the glut window
	//
	glutInitDisplayMode(GLUT_RGBA|GLUT_ALPHA|GLUT_DOUBLE|GLUT_DEPTH);
	glutInitWindowSize(outputWindowSize[0], outputWindowSize[1]);
	glutInitWindowPosition(100,100);
	main_window = glutCreateWindow("openGL5 - CS6610 - Justin Luitjens and Thiago Ize");

	
	// set callbacks
	glutDisplayFunc(myGlutDisplay);
	GLUI_Master.set_glutReshapeFunc(myGlutReshape);
	GLUI_Master.set_glutIdleFunc(myGlutIdle);
	GLUI_Master.set_glutKeyboardFunc(myGlutKeyboard);
	GLUI_Master.set_glutMouseFunc(myGlutMouse);
	glutMotionFunc(myGlutMotion);

	// create the interface subwindow and add widgets
	glui = GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_RIGHT);

	// initialize live variables
	currentTransforms = &modelTransform;

	lightTransform.live_object_xz_trans[0] = 5;
	lightTransform.live_object_xz_trans[1] = 2;
	lightTransform.live_object_y_trans = 5;

	hairVolumneTransform.live_object_xz_trans[0] = 0;
	hairVolumneTransform.live_object_xz_trans[1] = 0;
	hairVolumneTransform.live_object_y_trans = 1;

	// the object rollout
	object_rollout = glui->add_rollout("Object");

	// rotation and translation controls
	// we need an extra panel to keep things grouped properly
	GLUI_Panel *transform_panel = glui->add_panel_to_panel(object_rollout, "", GLUI_PANEL_NONE);
	object_rotation = glui->add_rotation_to_panel(transform_panel, "Rotation", currentTransforms->live_object_rotation);
	object_rotation->reset();

	glui->add_column_to_panel(transform_panel, false);
	object_xz_trans = glui->add_translation_to_panel(transform_panel, "Translate XZ", 
													 GLUI_TRANSLATION_XY, currentTransforms->live_object_xz_trans);

	glui->add_column_to_panel(transform_panel, false);
	object_y_trans =  glui->add_translation_to_panel(transform_panel, "Translate Y", 
													 GLUI_TRANSLATION_Y, &currentTransforms->live_object_y_trans);

	GLUI_Listbox *transformListbox = glui->add_listbox("Object to move", (int*)&currentTransforms, -1, switchTransform);
	transformListbox->add_item( (int)&modelTransform, "Model");
	transformListbox->add_item( (int)&lightTransform, "Light");
	transformListbox->add_item( (int)&hairVolumneTransform, "Hair Volume");
	

	object_xz_trans->scale_factor = 0.1f;
	object_y_trans->scale_factor = 0.1f;

	GLUI_Spinner *finHeightSpinner = glui->add_spinner("Fin Height", GLUI_SPINNER_FLOAT, &finHeight);
	finHeightSpinner->set_speed(.1);
	finHeightSpinner->set_float_limits(0,5);

	GLUI_Spinner *shellHeightSpinner = glui->add_spinner("Shell Height", GLUI_SPINNER_FLOAT, &shellHeight);
	shellHeightSpinner->set_speed(.1);
	shellHeightSpinner->set_float_limits(0,5);

	GLUI_Spinner *numHairSpinner = glui->add_spinner("numHairs", GLUI_SPINNER_INT, &numHairs, -1, makeHairGeometryVolume);
	numHairSpinner->set_speed(3);
	numHairSpinner->set_int_limits(1,10000);

	GLUI_Spinner *finScalingSpinner = glui->add_spinner("Fin Scaling Factor", GLUI_SPINNER_FLOAT, &finScalingFactor);
	finScalingSpinner->set_float_limits(.1,40);
	finScalingSpinner->set_speed(10);

	GLUI_Spinner *texture_scale = glui->add_spinner("Shell Texture Scale", GLUI_SPINNER_FLOAT, &tex_scale);
	texture_scale->set_float_limits(.1,20);
	texture_scale->set_speed(10);

	GLUI_Spinner *numShellsgSpinner = glui->add_spinner("Number of Shells", GLUI_SPINNER_INT, &numShells, -1, makeShells);
	numShellsgSpinner->set_int_limits(0, 100);

	glui->add_checkbox("Alpha Blend Fins", &blendFins);

	glui->add_checkbox("Draw Shell Box", &draw_shell_box);
	glui->add_checkbox("Draw Fin Box", &draw_fin_box);
	glui->add_checkbox("Draw Shells", &draw_shells);
	glui->add_checkbox("Draw Fins", &draw_fins);

	Hair::R=1;
	Hair::G=.6;
	Hair::B=0;
	
	glui->add_statictext("Hair Color");

	GLUI_Spinner *hair_color_red = glui->add_spinner("Red",GLUI_SPINNER_FLOAT,&Hair::R,0,makeHairUpdate);
	hair_color_red->set_float_limits(0,1);

	GLUI_Spinner *hair_color_green = glui->add_spinner("Green",GLUI_SPINNER_FLOAT,&Hair::G,0,makeHairUpdate);
	hair_color_green->set_float_limits(0,1);

	GLUI_Spinner *hair_color_blue = glui->add_spinner("Blue",GLUI_SPINNER_FLOAT,&Hair::B,0,makeHairUpdate);
	hair_color_blue->set_float_limits(0,1);

	glui->add_statictext("Skin Color");

	GLUI_Spinner *skin_color_red = glui->add_spinner("Red",GLUI_SPINNER_FLOAT,&skin_r,0,makeHairUpdate);
	skin_color_red->set_float_limits(0,1);

	GLUI_Spinner *skin_color_green = glui->add_spinner("Green",GLUI_SPINNER_FLOAT,&skin_g,0,makeHairUpdate);
	skin_color_green->set_float_limits(0,1);

	GLUI_Spinner *skin_color_blue = glui->add_spinner("Blue",GLUI_SPINNER_FLOAT,&skin_b,0,makeHairUpdate);
	skin_color_blue->set_float_limits(0,1);

	GLUI_Spinner *minShadowSpinner = glui->add_spinner("min Shadow",GLUI_SPINNER_FLOAT,&minShadow);
	minShadowSpinner->set_float_limits(0,1);

	GLUI_Spinner *fin_density= glui->add_spinner("Fin Density",GLUI_SPINNER_FLOAT,&finDensity,(int)&finDensity,makeHairUpdate);
	fin_density->set_float_limits(0,1);

	GLUI_Spinner *PdSpiner= glui->add_spinner("Pd",GLUI_SPINNER_FLOAT,&Pd,0,makeIlluminationTex);
	PdSpiner->set_float_limits(0,2000);
	PdSpiner->set_speed(.3);

	GLUI_Spinner *PsSpiner= glui->add_spinner("Ps",GLUI_SPINNER_FLOAT,&Ps,0,makeIlluminationTex);
	PsSpiner->set_float_limits(0,2000);
	PsSpiner->set_speed(.3);


		// empty space
	glui->add_statictext("");

	glui->set_main_gfx_window(main_window);

/*	
	GLenum err = glewInit();
	if (err != GLEW_OK) {
		printf("trouble with glewInit -- exiting\n");
		exit(-1);
	}
*/	

	// initialize gl
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	makeIlluminationTex(0);
	initializeCg();
	makeHairGeometryVolume(0);

	makeModel();
	makeDisplayList();
	// give control over to glut
	glutMainLoop();
	return 0;
}
